# ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: #
# Date: 25.09.2016                                                            #
# Author: Ole-Johan Skrede                                                    #
#                                                                             #
# Solution proposal as part of the exercise program in                        #
# INF4300 - Digital image analysis at the University of Oslo                  #
#                                                                             #
# ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: #
"""
Methods related to bit quad pattern matching.

For reference, see

Stephen B. Gray, 1971
  Local properties of binary images in two dimensions

Yang, Albregtsen, Lønnestad, Grøttum, 1994
  Methods to estimate areas and perimeters of blob-like objects: a comparison
"""

# pylint: disable=bad-indentation
# pylint: disable=too-many-return-statements

import numpy as np
import cv2

def get_patterns(name):
  """Return list of patterns (bit quads) associated with name"""

  if name == 'pixel':
    return [np.array([[1]])]
  elif name == 'hor':
    return [np.array([[0, 1]])]
  elif name == 'ver':
    return [np.array([[0], [1]])]
  elif name == 'Q0':
    return [np.array([[0, 0], [0, 0]])]
  elif name == 'Q1':
    return [np.array([[1, 0], [0, 0]]),
            np.array([[0, 1], [0, 0]]),
            np.array([[0, 0], [0, 1]]),
            np.array([[0, 0], [1, 0]])]
  elif name == 'Q2':
    return [np.array([[1, 1], [0, 0]]),
            np.array([[0, 1], [0, 1]]),
            np.array([[0, 0], [1, 1]]),
            np.array([[1, 0], [1, 0]])]
  elif name == 'Q3':
    return [np.array([[1, 1], [0, 1]]),
            np.array([[0, 1], [1, 1]]),
            np.array([[1, 0], [1, 1]]),
            np.array([[1, 1], [1, 0]])]
  elif name == 'Q4':
    return [np.array([[1, 1], [1, 1]])]
  elif name == 'QD':
    return [np.array([[1, 0], [0, 1]]),
            np.array([[0, 1], [1, 0]])]
  else:
    print('{} is not a valid name'.format(name))

def match_pattern(local_window, pattern):
  """Convenience function to match local_window with pattern. Allows local
  window to be interpreted as a binary image with gray levels in {0, 1}, by
  defining all values > 0 to be 1.
  """
  local_window = 1*(local_window > 0)

  if (local_window == pattern).all():
    return 1
  else:
    return 0

def count_matches(image, solid=False, seed_point=None, interior=False):
  """Count number of matches in a binary image.

  Args:
    image: graylevel image contating a foreground region to be examined.
    solid: bool. Whether the region is solid or only outlined.
    seed_point: bool. If the region is only outlined, specify seed point to
                fill from.
    interior: bool. If the region is only outlined, determine if the counting
              should include or exclude the outline.
  """

  # Convert from grayscale image to binary
  image = 1*(image > 0)

  height, width = image.shape
  if not solid:
    # Fill object inside contour with fill_value
    if interior:
      fill_value = 2
    else:
      fill_value = 1
    if seed_point is None:
      seed_point = (height//2, width//2)
    mask = np.zeros((height+2, width+2), np.uint8)
    flooded = image.astype(np.uint8)
    cv2.floodFill(flooded, mask, seed_point, fill_value)
    image = 1*(flooded == fill_value)

  num_matches = {}
  num_matches['pixel'] = 0
  num_matches['hor'] = 0
  num_matches['ver'] = 0
  num_matches['Q0'] = 0
  num_matches['Q1'] = 0
  num_matches['Q2'] = 0
  num_matches['Q3'] = 0
  num_matches['Q4'] = 0
  num_matches['QD'] = 0

  height, width = image.shape
  for i in range(height - 1):
    for j in range(width - 1):
      pixel_window = image[i:i+1, j:j+1]
      hor_window = image[i:i+1, j:j+2]
      ver_window = image[i:i+2, j:j+1]
      quad_window = image[i:i+2, j:j+2]

      for key, _ in num_matches.items():
        pattern_list = get_patterns(key)
        if key == 'pixel':
          local_window = pixel_window
        elif key == 'hor':
          local_window = hor_window
        elif key == 'ver':
          local_window = ver_window
        else:
          local_window = quad_window
        for pattern in pattern_list:
          num_matches[key] += match_pattern(local_window, pattern)

  return num_matches
